//
// vfive: Vector Field Interactive Visualization Environment
//
// basic.c++      Basic functions
//
//                by Kageyama, Akira  (kage@nifs.ac.jp)  2000.03.16
//                Modified by ohno 2008.3.10
//
// Please read the file "LICENCE_ja.txt"(in Japanese) or "LICENCE_en.txt" (in English)
// before you use this software. 

#include "vfive.h"

extern scalefHandle *scalefh;

//////////////////////////////////////////
// Utilities
//////////////////////////////////////////

void v5_error(char *last_will)
{
    printf("*** V5 ERROR *** %s\n", last_will);
    exit(1);
}

/*
 *-------------------------------------------------
 > v5_get_string
 *-------------------------------------------------
 *\

   Get a string from a VFIVE input data file by a key word.
   Here is a sample of the input data file.

   % cat sample.v5

   #
   # Sample file 
   # Comment begins with # or ' ' or '\n'.
   # keyword and filename is separated by blank(s).
    
   xfile  /tmp/sgks-x.dat
   yfile  /tmp/sgks-y.dat
   zfile  /tmp/sgks-z.dat
   cfile  /tmp/sgks-c.dat
    this  /is/comment, too, since it begins with a blank.

 */

void v5_get_string(char *file, char *the_key, char *the_name)
{
    FILE *fp;
    char line[200], name[160], key[40];
    fp = fopen(file, "r");
//	printf("debug %s\n", file);
    if (!fp)
	v5_error(" *** open failed. v5_get_string");

    while (fgets(line, 100, fp) != NULL) {
	if (line[0] == '#')
	    continue;		/* comment */
	if (line[0] == ' ')
	    continue;		/* comment */
	if (line[0] == '\n')
	    continue;		/* comment */
	int ln = (int)strlen(line);
	line[ln - 1] = '\0';	/* Put off the return code */
	sscanf(line, "%s %s", key, name);
	if (strcmp(key, the_key) == 0) {
	    strcpy(the_name, name);
	    break;
	}
    }

    fclose(fp);
}


/*
 *-------------------------------------------------
 > v5_get_int
 *-------------------------------------------------
 *
 *    To get integer from input.vpp.
 *
 */

void v5_get_int(char *file, char *the_key, int *ival)
{
    char s_ival[200];
    v5_get_string(file, the_key, s_ival);
    *ival = atoi(s_ival);
    printf("v5: %s --> %s = %d\n", file, the_key, *ival);
}


/*
 *-------------------------------------------------
 > v5_get_float
 *-------------------------------------------------
 *
 *    To get float precesion number from input.vpp.
 *      Added on 2008.05.27.
 *
 */

void v5_get_float(char *file, char *the_key, float *val)
{
    char s_val[200];
    v5_get_string(file, the_key, s_val);
    sscanf(s_val, "%f", val);
    printf("v5: %s --> %s = %f\n", file, the_key, *val);
}


/*
 *-------------------------------------------------
 > v5_get_double
 *-------------------------------------------------
 *
 *    To get double precesion number from input.vpp.
 *
 */

void v5_get_double(char *file, char *the_key, double *val)
{
    char s_val[200];
    v5_get_string(file, the_key, s_val);
    sscanf(s_val, "%lf", val);
    printf("v5: %s --> %s = %f\n", file, the_key, *val);
}



/*
 *---------------------------------------------------------------------
 > hsv2rgb
 *---------------------------------------------------------------------
 *
 *  Color code conversion. I forgot where I got the
 *  original code of this.      A. Kageyama
 *
 *
 */
void hsv2rgb(float h, float s, float v, float *r, float *g, float *b)
{
    int Hue;
    float Local, Diff, m, n, k;

    Local = h * 360.0;
    while (Local >= 360.0)
	Local = Local - 360.0;
    Local = Local / 60.0;
    Hue = (int) Local;
    Diff = Local - Hue;
    m = v * (1.0 - s);
    n = v * (1.0 - s * Diff);
    k = v * (1.0 - s * (1.0 - Diff));

    switch (Hue) {
    case 0:
	*r = v;
	*g = k;
	*b = m;
	break;
    case 1:
	*r = n;
	*g = v;
	*b = m;
	break;
    case 2:
	*r = m;
	*g = v;
	*b = k;
	break;
    case 3:
	*r = m;
	*g = n;
	*b = v;
	break;
    case 4:
	*r = k;
	*g = m;
	*b = v;
	break;
    case 5:
	*r = v;
	*g = m;
	*b = n;
    }

}

//////////////////////////////////////////
// Math
//////////////////////////////////////////

float vec_norm(float *v)
{
    return (sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]));
}

double vec_norm(double *v)
{
    return (sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]));
}


float vec_cross_normalf(float *a, float *b, float *c)
{
    c[0] = a[1] * b[2] - a[2] * b[1];	/* vector outer product */
    c[1] = a[2] * b[0] - a[0] * b[2];
    c[2] = a[0] * b[1] - a[1] * b[0];

    float nrm = vec_norm(c);

    c[0] /= nrm;
    c[1] /= nrm;
    c[2] /= nrm;

    return (nrm);
}


void vec_rot_z(const float in[], float out[], const float phi)
{
    float x = in[0];
    float y = in[1];

    out[0] = cos(phi) * x - sin(phi) * y;
    out[1] = sin(phi) * x + cos(phi) * y;
    out[2] = in[2];
}

float vec_amp(float v[])
{
    return (sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]));
}

void vec_normalize(float vec[])
{
    float vv;
    vv = vec_amp(vec);
    if (vv < 1.e-30) {
	puts("***  0 vector couldn't be converted into an unit vector.\n");
	puts("***  return the input vector and continue the job.\n");
    }
    vec[0] /= vv;
    vec[1] /= vv;
    vec[2] /= vv;
}

float vec_dotprdct(const float a[], const float b[])
{
    return (a[0] * b[0] + a[1] * b[1] + a[2] * b[2]);
}

void vec_f2ff(int n, ffloat_ out[], const float in[])
{
    for (int i = 0; i < n; i++)
	out[i] = in[i];
}

void vec_ff2f(int n, float out[], const ffloat_ in[])
{
    for (int i = 0; i < n; i++)
	out[i] = in[i];
}

void vec_f2f(int n, float out[], const float in[])
{
    for (int i = 0; i < n; i++)
	out[i] = in[i];
}


///////////////////////////////////////
// Dynamic array memory alloc
///////////////////////////////////////


/*
 *-------------------------------------------------
 > create_array_i1
 *-------------------------------------------------
 *
 *    To create    int a[ll]
 *
 */

int *create_array_i1(int ll)
{
    int *a;
    a = (int *) malloc(ll * sizeof(int));
    if (!a)
	v5_error("create_array_i1.");

    return (a);
}




/*
 *-------------------------------------------------
 > create_array_ff1
 *-------------------------------------------------
 *
 *    To create    ffloat_ a[ll]
 *
 */

ffloat_ *create_array_ff1(int ll)
{
    ffloat_ *a;
    a = (ffloat_ *) malloc(ll * sizeof(ffloat_));
    if (!a)
	v5_error("create_array_ff1.");

    return (a);
}



/*
 *-------------------------------------------------
 > create_array_ff2
 *-------------------------------------------------
 *
 *    To create    ffloat_ a[ll][mm]
 *
 */

ffloat_ **create_array_ff2(int ll, int mm)
{
    ffloat_ **a;
    ffloat_ *ptr;

    a = (ffloat_ **) malloc(ll * sizeof(ffloat_ *));
    if (!a)
	v5_error("create_array_ff2: 1");

    ptr = (ffloat_ *) malloc(ll * mm * sizeof(ffloat_));
    if (!ptr)
	v5_error("create_array_ff2: 2");


    for (int i = 0; i < ll; i++) {
	a[i] = ptr + i * mm;
    }

    return (a);
}



/*
 *-------------------------------------------------
 > create_array_ff3
 *-------------------------------------------------
 *
 *    To create    ffloat_ a[ll][mm][nn]
 *
 */

ffloat_ ***create_array_ff3(int ll, int mm, int nn)
{

    int i, j;

    ffloat_ ***a;
    ffloat_ *ptr;

    a = (ffloat_ ***) malloc(ll * sizeof(ffloat_ **));
    if (!a)
	v5_error("create_array_ff3: 1");

    for (i = 0; i < ll; i++) {
	a[i] = (ffloat_ **) malloc(mm * sizeof(ffloat_ *));
	if (!a[i])
	    v5_error("create_array_ff3: 2");
    }

    ptr = (ffloat_ *) malloc(ll * mm * nn * sizeof(ffloat_));
    if (!ptr)
	v5_error("create_array_ff3: 3");

    for (i = 0; i < ll; i++) {
	for (j = 0; j < mm; j++) {
	    a[i][j] = ptr + nn * (j + mm * i);
	}
    }

    return (a);
}



/*
 *-------------------------------------------------
 > create_array_ff4
 *-------------------------------------------------
 *
 *    To create    ffloat_ a[ll][mm][nn][oo]
 *
 */

ffloat_ ****create_array_ff4(int ll, int mm, int nn, int oo)
{

    int i, j, k;

    ffloat_ ****a;
    ffloat_ *ptr;

    a = (ffloat_ ****) malloc(ll * sizeof(ffloat_ ***));
    if (!a)
	v5_error("create_array_ff4: 1");

    for (i = 0; i < ll; i++) {
	a[i] = (ffloat_ ***) malloc(mm * sizeof(ffloat_ **));
	if (!a[i])
	    v5_error("create_array_ff4: 2");
	for (j = 0; j < mm; j++) {
	    a[i][j] = (ffloat_ **) malloc(nn * sizeof(ffloat_ *));
	    if (!a[i][j])
		v5_error("create_array_ff4: 3");
	}
    }

    ptr = (ffloat_ *) malloc(ll * mm * nn * oo * sizeof(ffloat_));
    if (!ptr)
	v5_error("create_array_ff4: 4");

    for (i = 0; i < ll; i++) {
	for (j = 0; j < mm; j++) {
	    for (k = 0; k < nn; k++) {
		a[i][j][k] = ptr + oo * (k + nn * (j + mm * i));
	    }
	}
    }

    return (a);
}



/****************************************
        float version 
 ****************************************/



/*
 *-------------------------------------------------
 > create_array_f1
 *-------------------------------------------------
 *
 *    To create    float a[ll]
 *
 */

float *create_array_f1(int ll)
{

    float *a;

    a = (float *) malloc(ll * sizeof(float));
    if (!a)
	v5_error("create_array_f1.");

    return (a);
}



/*
 *-------------------------------------------------
 > create_array_f2
 *-------------------------------------------------
 *
 *    To create    float a[ll][mm]
 *
 */

float **create_array_f2(int ll, int mm)
{
    float **a;
    float *ptr;

    a = (float **) malloc(ll * sizeof(float *));
    if (!a)
	v5_error("create_array_f2: 1");

    ptr = (float *) malloc(ll * mm * sizeof(float));
    if (!ptr)
	v5_error("create_array_f2: 2");


    for (int i = 0; i < ll; i++) {
	a[i] = ptr + i * mm;
    }

    return (a);
}



/*
 *-------------------------------------------------
 > create_array_f3
 *-------------------------------------------------
 *
 *    To create    float a[ll][mm][nn]
 *
 */

float ***create_array_f3(int ll, int mm, int nn)
{

    int i, j;

    float ***a;
    float *ptr;

    a = (float ***) malloc(ll * sizeof(float **));
    if (!a)
	v5_error("create_array_f3: 1");

    for (i = 0; i < ll; i++) {
	a[i] = (float **) malloc(mm * sizeof(float *));
	if (!a[i])
	    v5_error("create_array_f3: 2");
    }

    ptr = (float *) malloc(ll * mm * nn * sizeof(float));
    if (!ptr)
	v5_error("create_array_f3: 3");

    for (i = 0; i < ll; i++) {
	for (j = 0; j < mm; j++) {
	    a[i][j] = ptr + nn * (j + mm * i);
	}
    }

    return (a);
}



/*
 *-------------------------------------------------
 > create_array_f4
 *-------------------------------------------------
 *
 *    To create    float a[ll][mm][nn][oo]
 *
 */

float ****create_array_f4(int ll, int mm, int nn, int oo)
{

    int i, j, k;

    float ****a;
    float *ptr;

    a = (float ****) malloc(ll * sizeof(float ***));
    if (!a)
	v5_error("create_array_f4: 1");

    for (i = 0; i < ll; i++) {
	a[i] = (float ***) malloc(mm * sizeof(float **));
	if (!a[i])
	    v5_error("create_array_f4: 2");
	for (j = 0; j < mm; j++) {
	    a[i][j] = (float **) malloc(nn * sizeof(float *));
	    if (!a[i][j])
		v5_error("create_array_f4: 3");
	}
    }

    ptr = (float *) malloc(ll * mm * nn * oo * sizeof(float));
    if (!ptr)
	v5_error("create_array_f4: 4");

    for (i = 0; i < ll; i++) {
	for (j = 0; j < mm; j++) {
	    for (k = 0; k < nn; k++) {
		a[i][j][k] = ptr + oo * (k + nn * (j + mm * i));
	    }
	}
    }

    return (a);
}

/******************** unsigned char version ****************/
/* by Nobuaki, Ohno 2004*/
unsigned char ****create_array_uc4(int ll, int mm, int nn, int oo)
{

    int i, j, k;

    unsigned char ****a;
    unsigned char *ptr;

    a = (unsigned char ****) malloc(ll * sizeof(unsigned char ***));
    if (!a)
	v5_error("create_array_uc4: 1");

    for (i = 0; i < ll; i++) {
	a[i] = (unsigned char ***) malloc(mm * sizeof(unsigned char **));
	if (!a[i])
	    v5_error("create_array_uc4: 2");
	for (j = 0; j < mm; j++) {
	    a[i][j] =
		(unsigned char **) malloc(nn * sizeof(unsigned char *));
	    if (!a[i][j])
		v5_error("create_array_uc4: 3");
	}
    }

    ptr =
	(unsigned char *) malloc(ll * mm * nn * oo *
				 sizeof(unsigned char));
    if (!ptr)
	v5_error("create_array_uc4: 4");

    for (i = 0; i < ll; i++) {
	for (j = 0; j < mm; j++) {
	    for (k = 0; k < nn; k++) {
		a[i][j][k] = ptr + oo * (k + nn * (j + mm * i));
	    }
	}
    }

    return (a);
}

float scaleintp()
{
    float scale;
    scale = scalefh->current_scale_is();

    return scale;

}

