//
// vfive: Vector Field Interactive Visualization Environment
//
//
// scalef_handle.c++
//
//                by Nobuaki, Ohno 2004.10.
//
// Please read the file "LICENCE_ja.txt"(in Japanese) or "LICENCE_en.txt" (in English)
// before you use this software. 

#include "vfive.h"

extern GLUquadricObj *cyl[CAVE_MAX_WALLS];                   // cylinder

scalefHandle::scalefHandle()
{
  vert_level_bar_rad_max = 0.2;
  vert_level_bar_height = 2.0;
  set_current_scale (1.0);
  on = 0;

  work_level = current_scale_is ();
  work_vmin = MIN_SCALE;
  work_vmax = MAX_SCALE;
  work_dval = work_vmax - work_vmin;
  work_aver = (MAX_SCALE - MIN_SCALE) / 2.0;
  work_vari =
    sqrt ((MAX_SCALE * MAX_SCALE * MAX_SCALE -
	   MIN_SCALE * MIN_SCALE * MIN_SCALE) / 2.0 - work_aver * work_aver);

  float abs_vmax = fabs (work_vmax);
  float abs_vmin = fabs (work_vmin);
  float abs_larger = abs_vmax > abs_vmin ? abs_vmax : abs_vmin;
  vert_drad_dlevel = vert_level_bar_rad_max / abs_larger;
}


void
scalefHandle::make_vert_level_bar ()
{
  GLfloat diffuse[] = { 0.3, 0.3, 1.0, 1.0 };
  GLfloat ambient[] = { 0.0, 0.3, 0.3, 1.0 };
  GLfloat specular[] = { 0.3, 1.0, 0.3, 1.0 };

  glEnable (GL_LIGHTING);
  glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);
  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
  glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, 10.0);

  float rad_top = vert_level_bar_rad (work_vmax);
  float rad_btm = vert_level_bar_rad (work_vmin);
  int tID = CAVEUniqueIndex();

  glPushMatrix ();
  if (work_vmax * work_vmin > 0.0)
    {				// Same sign. A cone.
      gluCylinder (cyl[tID], rad_btm, rad_top, vert_level_bar_height, 16, 2);
    }
  else
    {				// Change sign.
      float ratio_plus = work_vmax / work_dval;
      float ratio_minus = 1.0 - ratio_plus;
      float height_plus = vert_level_bar_height * ratio_plus;
      float height_minus = vert_level_bar_height * ratio_minus;
      gluCylinder (cyl[tID], rad_btm, 0.0, height_minus, 16, 2);
      glTranslatef (0.0, 0.0, height_minus);
      gluCylinder (cyl[tID], 0.0, rad_top, height_plus, 16, 2);
    }
  glPopMatrix ();
}

static void
draw_circle (float rad)
{
  const int N = 60;
  const float dtht = float (TWOPI) / N;
  glBegin (GL_LINE_LOOP);
  for (int i = 0; i < N; i++)
    {
      float tht = dtht * i;
      float x = rad * cos (tht);
      float y = rad * sin (tht);
      glVertex3f (x, y, 0);
    }
  glEnd ();
}

void
scalefHandle::ring (float lev)
{
  float dzdl = vert_level_bar_height / work_dval;
  float z = dzdl * (lev - work_vmin);
  float r = vert_level_bar_rad (lev) + 0.01;	// 0.01 feet away

  glPushMatrix ();
  glTranslatef (0.0, 0.0, z);
  draw_circle (r);
  glPopMatrix ();
}


void
scalefHandle::make_level_rings ()
{
  float rad_top = vert_level_bar_rad (work_vmax);
  float rad_btm = vert_level_bar_rad (work_vmin);

  glDisable (GL_LIGHTING);

  float lev = work_aver + work_vari;

  // val = average
  glColor3f (0.0, 0.0, 0.0);
  lev = work_aver;
  ring (lev);

  // val > average
  glColor3f (1.0, 0.0, 1.0);
  while (lev < work_vmax)
    {
      ring (lev);
      lev += work_vari;
    }

  // val < average
  lev = work_aver - work_vari;
  while (lev > work_vmin)
    {
      ring (lev);
      lev -= work_vari;
    }

}


static void
arrow_color ()
{
  GLfloat diffuse[] = { 0.8, 0.8, 0.3, 1.0 };
  GLfloat ambient[] = { 0.3, 0.3, 0.0, 1.0 };
  GLfloat specular[] = { 1.0, 1.0, 0.3, 1.0 };

  glMaterialfv (GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);
  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
  glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, 10.0);
}

void
scalefHandle::left_arrow ()
  //    
  //     ||
  //     || /
  //     ||<  Make an left arrow which touches the vertical bar.
  //     || \
  //     ||
  //
{
  float rad = 0.1;
  float length = 0.5;
  glTranslatef (length + vert_level_bar_rad (work_level), 0.0, 0.0);
  glRotatef (-90.0, 0.0, 1.0, 0.0);
  gluCylinder (cyl[CAVEUniqueIndex()], rad, 0.0, length, 16, 2);
}

void
scalefHandle::all_arrows ()
{
  int n_arrow = 8;
  float dtht = inl_r2d (TWOPI / n_arrow);
  glEnable (GL_LIGHTING);
  arrow_color ();
  for (int i = 0; i < n_arrow; i++)
    {
      glPushMatrix ();
      glRotatef (dtht * i, 0.0, 0.0, 1.0);
      left_arrow ();
      glPopMatrix ();
    }
}

void
scalefHandle::wand_motion ()
{
  float pos[3];
  float dz;

  CAVEGetPosition (CAVE_WAND_NAV, pos);

  dz = pos[2] - snapstate.wand_pos_z ();

  float ratio = dz / vert_level_bar_height;
  float old_level = current_scale_is ();
  float new_level = old_level + work_dval * ratio;
  if (new_level > work_vmax)
    new_level = work_vmax;
  if (new_level < work_vmin)
    new_level = work_vmin;
  work_level = new_level;
}

static float
rot_angle ()
{
  const float dphidt = 36.0;	// 1 rotation in 10 seconds.
  float t = CAVEGetTime ();
  return (dphidt * t);
}

void
scalefHandle::draw_vertical_bar (float x, float y, float z,
				 float phi, float distance)
{
  //     
  //       ||
  //       || The radius of bar changes acoord to scalar value.
  //       ||
  //       ||
  //       ||
  //   --> || <--   level indicator by 8 arrows. They slides by wand.
  //       ||
  //       ||
  //       ||
  //

  x -= distance * sin (phi);	// Since phi=0 is y-axis.
  y += distance * cos (phi);
  z -= 1.5;			// Shift 0.5 feet down.

  float dlev = (work_level - work_vmin) / work_dval;
  float shift_z = dlev * vert_level_bar_height;

  glPushAttrib (GL_LIGHTING_BIT);
  glPushMatrix ();
  glTranslatef (x, y, z);
  make_vert_level_bar ();
  make_level_rings ();
  glPushMatrix ();
  glRotatef (rot_angle (), 0.0, 0.0, 1.0);	// Rotation around the bar.
  glTranslatef (0.0, 0.0, shift_z);
  all_arrows ();
  glPopMatrix ();
  glPopMatrix ();
  glPopAttrib ();
}


void
scalefHandle::draw ()
{
  float scale;

  if (!on)
    return;

  scale = current_scale_is ();

  float x = snapstate.head_pos_x ();
  float y = snapstate.head_pos_y ();
  float z = snapstate.head_pos_z ();
  float distance = 3.0;		// feet away
  float phi = snapstate.head_phi () + inl_d2r (20.0);	// slightly left.

  glScalef (1.0 / scale, 1.0 / scale, 1.0 / scale);

  draw_vertical_bar (x, y, z, phi, distance);

  glScalef (scale, scale, scale);


}


void
scalefHandle::snap ()
{
  snapstate.record_head_info ();	// When middle button is pressed.
  snapstate.record_wand_info ();	// Wand is used to shift the cut plane.
}

void
scalefHandle::button_has_pressed ()
{
  snap ();

  work_level = current_scale_is ();
  work_vmin = MIN_SCALE;
  work_vmax = MAX_SCALE;
  work_dval = work_vmax - work_vmin;
  work_aver = (MAX_SCALE - MIN_SCALE) / 2.0;
  work_vari =
    sqrt ((MAX_SCALE * MAX_SCALE * MAX_SCALE -
	   MIN_SCALE * MIN_SCALE * MIN_SCALE) / 2.0 - work_aver * work_aver);

  float abs_vmax = fabs (work_vmax);
  float abs_vmin = fabs (work_vmin);
  float abs_larger = abs_vmax > abs_vmin ? abs_vmax : abs_vmin;
  vert_drad_dlevel = vert_level_bar_rad_max / abs_larger;
  turn (1);
}

float
scalefHandle::vert_level_bar_rad (float level)
{
  return fabs (level) * vert_drad_dlevel;
}



void
scalefHandle::button_has_released ()
{
  set_current_scale (work_level);
  turn (0);
}

void
scalefHandle::button_being_pressed ()
{
  if(!on) return; //2008.4.21
  wand_motion ();
}

void
scalefHandle::button3_has_pressed ()
{
  set_current_scale (1.0);
}
