//
// vfive: Vector Field Interactive Visualization Environment
//
// snow.c++
//
//                by Kageyama, Akira  (kage@nifs.ac.jp)  2000.04.18
//  
// Please read the file "LICENCE_ja.txt"(in Japanese) or "LICENCE_en.txt" (in English)
// before you use this software. 


#include "vfive.h"

extern Data *data;

Snow::Snow(float spot_distance_, float spot_angle_)
{
  spot_distance = spot_distance_;
  spot_angle = inl_d2r(spot_angle_);
  on = 0;
  integration = 0;
  svec_ = 0;

  for (int i=0; i<SNOW_NUM; i++){
   snow_pos[i][0] = 0.0;
   snow_pos[i][1] = 0.0;
   snow_pos[i][2] = 0.0;

   snow_col[i][0] = 0.0;
   snow_col[i][1] = 0.0;
   snow_col[i][2] = 0.0;
  }

  for (int i=0; i<3; i++){
   wand_pos[i] = 0.0;     // These four vectors are used 
   wand_dir[i] = 0.0;     // to define the spot light cone.
   wand_rit[i] = 0.0;  
   wand_up[i]  = 0.0;
  }

  refresh_spot = 0;

  for (int i=0; i<SNOW_NUM; i++)
    reposit_snow(i);
}

static double random0to1(){

    const int N1=1229, N2=351750, N3=1664501;
    static int nrandm = 229933;//any integer
    double arandm;

    nrandm = (N1*nrandm+N2)%N3;
    arandm = (double)nrandm/(double)N3;

	return arandm;
}

int
Snow::is_in_spot_light(float *r_vec) 
{
  //
  // Check if the snow particle is inside (1) of the cone region
  // of the spot light emitted from the wand, or not (0).
  // The cone region is defined by the length 'spot_distance'
  // and the angle 'spot_angle'.
  //

  float distance = vec_amp(r_vec);
  if (distance > spot_distance) return 0;    // Too far.

  float unit_v[3];

  for (int i=0; i<3; i++)                 // Normalization
    unit_v[i] = r_vec[i] / distance; 

  float theta = acos(vec_dotprdct(wand_dir, unit_v));

  if (theta <= spot_angle) return 1;      // OK. It's in the spot.
  return 0;                               // Outside of the spot.
}


float 
Snow::weighted_random_rsq()
  //
  // Returns "weighted random numbers".
  // Its distribution is proportional to x^2 where
  //        0 <= x < 1.0
  //       
{
    float r;
    do {
//      r = drand48();
      r = (float)random0to1();
      } while ( r*r <= (float)random0to1() );
    return r;
}

float 
Snow::weighted_random_sint()
  //
  // Returns "weighted random numbers".
  // Its distribution is proportional to sin(theta)
  //        and 0 for theta > spot_angle
  //       
{
    float t;
    float sin_t_critical = sin(spot_angle);
    do {
      t = (float)random0to1() * spot_angle;
    } while ( sin(t) <= (float)random0to1()*sin_t_critical );
    return t;
}

void
Snow::reposit_snow(int i)
  // 
  // Place a snow particle in the spot-light-cone
  // with uniform distribution density.
  //
{
  float rad = spot_distance * weighted_random_rsq(); // distrib prop to r^2.
  float theta = weighted_random_sint();        // prop to sin(theta).
  float phi = TWOPI * (float)random0to1();               // uniform in phi.
  //
  //         wand pos -----------------------> (0)
  //
  float x0 = wand_pos[0] + wand_dir[0] * rad;
  float y0 = wand_pos[1] + wand_dir[1] * rad;
  float z0 = wand_pos[2] + wand_dir[2] * rad;
  //
  //         wand pos -----------------------+ (0)
  //                                         |
  //                                         |
  //                                         o
  //                                        (1)
  float distance_0to1 = rad*sin(theta);
  float unit_v0to1_x = cos(phi)*wand_rit[0] + sin(phi)*wand_up[0];
  float unit_v0to1_y = cos(phi)*wand_rit[1] + sin(phi)*wand_up[1];
  float unit_v0to1_z = cos(phi)*wand_rit[2] + sin(phi)*wand_up[2];

  // Finally...
  float new_x = x0 + distance_0to1*unit_v0to1_x;
  float new_y = y0 + distance_0to1*unit_v0to1_y;
  float new_z = z0 + distance_0to1*unit_v0to1_z;
  if (   new_x >= data->region_xmin && new_x <= data->region_xmax
      && new_y >= data->region_ymin && new_y <= data->region_ymax
      && new_z >= data->region_zmin && new_z <= data->region_zmax) {
    snow_pos[i][0] = new_x;
    snow_pos[i][1] = new_y;
    snow_pos[i][2] = new_z;
//	printf("snow_pos (%f,%f,%f) wand_pos(%f,%f,%f)\n",snow_pos[i][0],snow_pos[i][1],snow_pos[i][2],
//							wand_pos[0],wand_pos[1],wand_pos[2]);
  }
  else {
    snow_pos[i][0] = 9.99e19;  // Not shown since it's too far (probably.)
    snow_pos[i][1] = 9.99e19;
    snow_pos[i][2] = 9.99e19;
  }
}

void
Snow::initialize() 
{
//  for (int i=0; i<SNOW_NUM; i++)
//    reposit_snow(i);
}


void
Snow::integrate()
{
  if (!on) return;
  
  int id;
  float vec[3], new_pos[3], vec_wand2snow[3];
  float val;
  float dt = 0.1 / data->vector_fld[svec_]->vmax;
  
  for (int i=0; i<SNOW_NUM; i++) {
    id = intpol_get_vect_for_snow(snow_pos[i], vec, &val);
    if (id) {
      reposit_snow(i);
      continue;
    }
    new_pos[0] = snow_pos[i][0] + vec[0] * dt;
    new_pos[1] = snow_pos[i][1] + vec[1] * dt;
    new_pos[2] = snow_pos[i][2] + vec[2] * dt;

	vec_wand2snow[0] = new_pos[0] - wand_pos[0];
    vec_wand2snow[1] = new_pos[1] - wand_pos[1];
    vec_wand2snow[2] = new_pos[2] - wand_pos[2];
    if (is_in_spot_light(vec_wand2snow) ) {
      snow_pos[i][0] = new_pos[0];
      snow_pos[i][1] = new_pos[1];
      snow_pos[i][2] = new_pos[2];
    }
    else {
      reposit_snow(i);
      continue;
    }
  }

    for (int i = 0; i < SNOW_NUM; i++) {
	float dist = sqrt((snow_pos[i][0]-wand_pos[0])*(snow_pos[i][0]-wand_pos[0])+
	(snow_pos[i][1]-wand_pos[1])*(snow_pos[i][1]-wand_pos[1]) +
(snow_pos[i][2]-wand_pos[2])*(snow_pos[i][2]-wand_pos[2]) );

	float kido = (13.0 - dist)/10.0;
	if(kido > 1.0) kido = 1.0;
		else if(kido <=0.0) kido = 0.0;

	snow_col[i][0] = kido;
	snow_col[i][1] = kido;
	snow_col[i][2] = kido;

    }

}

void
Snow::set_spot_light()
{
  float scale;
  CAVEGetPosition(CAVE_WAND_NAV,   wand_pos);
  CAVEGetVector(CAVE_WAND_FRONT_NAV, wand_dir);
  CAVEGetVector(CAVE_WAND_RIGHT_NAV, wand_rit);
  CAVEGetVector(CAVE_WAND_UP_NAV,    wand_up);

  scale = scaleintp();
  wand_pos[0] = wand_pos[0]/scale;
  wand_pos[1] = wand_pos[1]/scale;
  wand_pos[2] = wand_pos[2]/scale;
}


void
Snow::draw()
{
  if (!on) return;             // Now it's off.

  glDisable(GL_LIGHTING); 
  glPushAttrib(GL_POINT);
  glPointSize(1.5);
  glBegin(GL_POINTS);
  glColor3f(1.0, 1.0, 1.0);
  for (int i=0; i<SNOW_NUM; i++) {
    glColor3fv(snow_col[i]);
    glVertex3fv(snow_pos[i]);
  }
  glEnd();
  glPopAttrib();
}

void
Snow::turn(int i)
{
  if (i == 1) {
    on = 1;
//    set_spot_light();
  }
  else {
    on = 0;
    refresh_spot = 0;
  }
}

void
Snow::button_being_pressed()
{
  if(!on) return;//2008.4.21
  set_spot_light();
}

void
Snow::button_has_pressed()
{
  refresh_spot = 1;
  set_spot_light(); //2008.4.21
}

void
Snow::button3_has_pressed()
{
  turn(0);
}


void
Snow::button_has_released()
{
  refresh_spot = 0;
}

