//
// vfive: Vector Field Interactive Visualization Environment
//
// tracer.c++
//
//                by Kageyama, Akira  (kage@nifs.ac.jp)  2000.03.17
//  
// Please read the file "LICENCE_ja.txt"(in Japanese) or "LICENCE_en.txt" (in English)
// before you use this software. 


#include "vfive.h"

extern Data *data;
extern GLUquadricObj *sphereObj[CAVE_MAX_WALLS];             // Ball

static
tracer_line_ptr**
create_array_2D_of_tracer_line_ptr(int ll, int mm)
{
    tracer_line_ptr **a;
    tracer_line_ptr *ptr;

    a = (tracer_line_ptr **)malloc(ll * sizeof(tracer_line_ptr *));
    if (!a) v5_error("create_array_2D_of_tracer_line_ptr: 1");

    ptr = (tracer_line_ptr *)malloc(ll * mm * sizeof(tracer_line_ptr));
    if (!ptr) v5_error("create_array_2D_of_tracer_line_ptr: 2");

    for (int i=0; i<ll; i++){
        a[i] = ptr +  i * mm;
    }

    return(a);

}


Tracer::Tracer()
{
  int i;
  turn_off();
  integration_off();
  svec_ = 0;

  alin = create_array_i1(data->nvec);

  line = create_array_2D_of_tracer_line_ptr(data->nvec, TRACER_LINE_MAX);

  for (i=0; i<data->nvec; i++) {
    for (int j=0; j<TRACER_LINE_MAX; j++) {
      line[i][j] = new Tracer_line;
      line[i][j]->set_svec(i);
       if(line[i][j] == NULL) v5_error("can't allocate tracer_line: tracer");
    }
  }

  beam = new PointerBeam(0.3, 1.0, 1.0, 3.0);
  if(beam == NULL) v5_error("can't allocate beam: tracer");

  for (int n=0; n<data->nvec; n++) 
    alin[n] = 0;

  Color wrk;
  i = 0;
  while (i < data->nvec) {
    vec_f2f(4, coltable[i++], wrk.mizuiro(0.0));
    if (i>=MAX_NVEC) break;
    vec_f2f(4, coltable[i++], wrk.aka(0.6));     // pink.
    if (i>=MAX_NVEC) break;
    vec_f2f(4, coltable[i++], wrk.midori(0.5));  // light green.
    if (i>=MAX_NVEC) break;
    vec_f2f(4, coltable[i++], wrk.murasaki(0.3));  
    if (i>=MAX_NVEC) break;
  }

}


static void
next_line(int& line)
{
  line = (++line) % TRACER_LINE_MAX;
}


void
Tracer::release()
{
  int sv = svec();

  switch (direct) {
  case POSI:
    line[sv][alin[sv]]->erase();                  // delete the old line.
    line[sv][alin[sv]]->set_trace_normed_vec(0);  // Trace raw vector.
    line[sv][alin[sv]]->start(beam, sv, +1, coltable[sv]); // positive trace
    next_line(alin[sv]);
    break;

  case NEGA:
    line[sv][alin[sv]]->erase();                  // delete the old line.
    line[sv][alin[sv]]->set_trace_normed_vec(0);  // Trace raw vector.
    line[sv][alin[sv]]->start(beam, sv, -1, coltable[sv]); // negative trace
    next_line(alin[sv]);
    break;

  case BOTH:
    line[sv][alin[sv]]->erase();                  // delete the old line.
    line[sv][alin[sv]]->set_trace_normed_vec(1);  // Trace raw vector.
    line[sv][alin[sv]]->start(beam, sv, +1, coltable[sv]); // positive trace
    next_line(alin[sv]);
    line[sv][alin[sv]]->erase();                  // delete the old line.
    line[sv][alin[sv]]->set_trace_normed_vec(1);  // Trace raw vector.
    line[sv][alin[sv]]->start(beam, sv, -1, coltable[sv]); // negative trace
    next_line(alin[sv]);
    break;
  }
}

void
Tracer::integrate()
{
  if (!integration) return;

  for (int v=0; v<data->nvec; v++) {
    for (int i=0; i<TRACER_LINE_MAX; i++) {
      if (line[v][i]->calc_state()) line[v][i]->integrate();
    }
  }
}

void 
Tracer::draw_orbits(void)
{
  glLineWidth(1.0); // originary 2.0
  for (int v = 0; v < data->nvec; v++) {
    glColor3fv(coltable[v]);
    for (int i = 0; i < TRACER_LINE_MAX; i++) {
      if (line[v][i]->on == 0) continue;
      int n = line[v][i]->nten;
      glPushMatrix();
      glBegin(GL_LINE_STRIP);
        // for (j=0; j<n; j+=20)    fast drawing by skip 
        for (int j=0; j<n; j++)
	  glVertex3fv(line[v][i]->verts[j]);
      glEnd();
      glPopMatrix();
    }
  }
}

void
Tracer::reset()
{

  int v = svec();
  alin[v] = 0;
  for (int i = 0; i < TRACER_LINE_MAX; i++) 
    line[v][i]->erase();
}


void 
Tracer::draw_particles()
{
  int tID;
  tID = CAVEUniqueIndex();

  for (int v=0; v < data->nvec; v++) {
    glMaterialfv(GL_FRONT, GL_DIFFUSE, coltable[v]);
    for (int i=0; i < TRACER_LINE_MAX; i++) {
      if (line[v][i]->on == 0) continue;
      int n = line[v][i]->nten;
      float x = line[v][i]->verts[n][0];
      float y = line[v][i]->verts[n][1];
      float z = line[v][i]->verts[n][2];
      glPushMatrix();
        glTranslatef(x, y, z);   // tip of the orbit
        gluSphere(sphereObj[tID], line[v][i]->radius, 8, 8);
      glPopMatrix();
    }
  }

}

void
Tracer::initialize() 
{

}

void
Tracer::draw()
{
  if (!on) return;             // Now it's off.

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING); 
  beam->draw();
  draw_orbits();
  glEnable(GL_LIGHTING);
  draw_particles();
  glPopAttrib();
}

void
Tracer::button_being_pressed()
{
  if(!on) return;//2008.4.21
  beam->reset();
}

void
Tracer::button_has_pressed()
{
  beam->reset(); //2008.4.21
  beam->turn(1);
}

void
Tracer::button3_has_pressed()
{
  reset();
}


void
Tracer::button_has_released()
{
  beam->turn(0);
  release();      // Button release releases a ball!
}



