/////////////////////////////////////////////////////////////
// vfive: Vector Field Interactive Visualization Environment
//
// volren.c++      Volume Rendering (3D texture)
//
// by Nobuaki, Ohno  2008.12
/////////////////////////////////////////////////////////////

#include "vfive.h"

#ifdef WIN32
PFNGLTEXIMAGE3DPROC glTexImage3D;
#endif

extern Data *data;
extern scalefHandle *scalefh;
extern GLUquadricObj *sphereObj[CAVE_MAX_WALLS]; // Ball on the orbits' tip

static const float BALL_RAD = 0.05;

static const int SCALAR_MAX = 256;

static const float G_XORIGIN = 0.0;
static const float G_YORIGIN = 4.0;
static const float G_ZORIGIN = -5.0;

static const float GRAPH_SCALE = 0.01;

static const float X_GRAPH_LEN = (SCALAR_MAX * GRAPH_SCALE);
static const float Y_GRAPH_LEN = (SCALAR_MAX * GRAPH_SCALE);

static const float COLOR_BAR_HEIGHT = 0.24;

static const float SQ_LEN = X_GRAPH_LEN / 6.5;
static const float SQ_SPACE = (X_GRAPH_LEN - SQ_LEN * 5) / 4.0;
static const float SQ_YOFFSET = Y_GRAPH_LEN + 0.14 * 100.0 * GRAPH_SCALE;

///////////////////////////////
// Initialization
///////////////////////////////
cave_volrend::cave_volrend()
{

    float len1, len2, len3, slice_delta;

    on = 0;
    volumedata = create_array_uc4(data->nscal, IV3, IV2, IV1);
    setcolortable();
    dispmode = 0;
    reset = 0;
    select_scal(0);

    //color editor
    memset(&cedata, 0, sizeof(coloreditordata));
    cedata.element = 3;

    //volume rendering
    len1 = (data->region_xmax - data->region_xmin);
    len2 = (data->region_ymax - data->region_ymin);
    len3 = (data->region_zmax - data->region_zmin);

    ///////////////////
    // Slice no kankaku
    ///////////////////
    slice_delta = ((len1 / IV1 + len2 / IV2 + len3 / IV3) / 3.0) / 1.25;

    datasize[0] = len1;
    datasize[1] = len2;
    datasize[2] = len3;

    float dia = sqrt(len1 * len1 + len2 * len2 + len3 * len3);

    memset(&slicedata, 0, sizeof(slicedata));
    slicedata.slice_max_limit = (int) (dia / slice_delta + 10.0);// Max number of slices
    slicedata.delta = slice_delta;	// Slice no kankaku
    make_slicedata();


    double epsi = 1.e-6;
    double safty = 1.0 - epsi * 2;
    double xleng = (data->region_xmax - data->region_xmin) * safty;
    double yleng = (data->region_ymax - data->region_ymin) * safty;
    double zleng = (data->region_zmax - data->region_zmin) * safty;

    init_vertx(); //after datasize

    data_origin[0] = data->region_xmin + xleng / 2.0;
    data_origin[1] = data->region_ymin + yleng / 2.0;
    data_origin[2] = data->region_zmin + zleng / 2.0;

    ffloat_ x[IV1];
    ffloat_ y[IV2];
    ffloat_ z[IV3];
    ffloat_ value, vmin, vmax;
    int intensity;
    ffloat_ cx[4], cy[4], cz[4];
    int ii, jj, kk;
    ffloat_ dx = xleng / (IV1 - 1);
    ffloat_ dy = yleng / (IV2 - 1);
    ffloat_ dz = zleng / (IV3 - 1);

    for (int i = 0; i < IV1; i++)
	x[i] = data->region_xmin + epsi + dx * i;

    for (int j = 0; j < IV2; j++)
	y[j] = data->region_ymin + epsi + dy * j;

    for (int k = 0; k < IV3; k++)
	z[k] = data->region_zmin + epsi + dz * k;


    for (int n = 0; n < data->nscal; n++){

//     for (int n = 0; n < data->nscal; n++) {

	vmin = data->scalar_fld[n]->vmin;
	vmax = data->scalar_fld[n]->vmax;

	for (int i = 0; i < IV1; i++) {
	    for (int j = 0; j < IV2; j++) {
		for (int k = 0; k < IV3; k++) {
		    ffloat_ x_ = x[i];
		    ffloat_ y_ = y[j];
		    ffloat_ z_ = z[k];
		    int iout =
			find_fact(x_, y_, z_, &ii, &jj, &kk, cx, cy, cz);

		    if (iout) {
			v5_error("Volumerendering: out of box \n");
		    }

		    value =
			interpolator(ii, jj, kk, cx, cy, cz,
				     data->scalar_fld[n]->val);

		    intensity =
			(int) (256.0 * ((value - vmin) / (vmax - vmin)));

		    if (intensity < 0)
			intensity = 0;
		    if (intensity > 255)
			intensity = 255;

		    volumedata[n][k][j][i] = (unsigned char) intensity;

		}
	    }

       }
    }

    recalc_flag = 0;
}

// Memory Allocation
//////////////////////////////////////
void cave_volrend::make_slicedata()
{

    int n = slicedata.slice_max_limit;

    slicedata.angles = (int *) malloc(sizeof(int) * n);
    slicedata.slicevertx =
	(struct slicevertex *) malloc(sizeof(struct slicevertex) * n);

    memset(slicedata.angles, 0, sizeof(int) * n);
    memset(slicedata.slicevertx, 0, sizeof(struct slicevertex) * n);

}


/////////////////////////////////
///////////////////////////////////////////////////////////////////////////

void cave_volrend::clean()
{
    free(slicedata.angles);
    free(slicedata.slicevertx);

}

//////////////////////////////////////////////
// For Color Editor :called in wand_button.c++
//////////////////////////////////////////////
void cave_volrend::button_being_pressed()
{
    if (!on)
	return;

    cedata.ray_exist = 1;
    dispmode = 1;

    check_grab();
    update_held();

}


////////////////////////////////////////////
// Initialization
// ---------------
// Making Volume Data & Graph (Color Editor)
////////////////////////////////////////////
void cave_volrend::initialize()
{
    int tID = CAVEUniqueIndex();

#ifdef WIN32
  glTexImage3D =
    (PFNGLTEXIMAGE3DPROC)wglGetProcAddress("glTexImage3D");
#endif

    for (int n = 0; n < data->nscal; n++) {
	glGenTextures(1, &tex[tID][n]);
    }


    for (int n = 0; n < data->nscal; n++){
	CAVEDisplayBarrier();
	if(CAVEMasterDisplay()) remake_voldata(n);
	CAVEDisplayBarrier();
//    	recalc(n);
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glBindTexture(GL_TEXTURE_3D, tex[tID][n]);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA, IV1, IV2, IV3,
		     0, GL_RGBA, GL_UNSIGNED_BYTE,
		     &texvoldata[0][0][0]);

    }

    volren_graph[tID] = glGenLists(1);
    glNewList(volren_graph[tID], GL_COMPILE);
    glPushAttrib(GL_LINE_BIT);
    glBegin(GL_LINE_STRIP);
    glVertex3f(0, Y_GRAPH_LEN, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(X_GRAPH_LEN, 0, 0);
    glVertex3f(X_GRAPH_LEN, Y_GRAPH_LEN, 0);
    glVertex3f(0, Y_GRAPH_LEN, 0);
    glEnd();
    glPopAttrib();
    glEndList();

}

void cave_volrend::remake_voldata(int n){

    unsigned char intensity;

	for (int i = 0; i < IV1; i++) {
	    for (int j = 0; j < IV2; j++) {
		for (int k = 0; k < IV3; k++) {

		    intensity = volumedata[n][k][j][i];
		    texvoldata[k][j][i*4 + 0] = colortable[n][intensity][0];
		    texvoldata[k][j][i*4 + 1] = colortable[n][intensity][1];
		    texvoldata[k][j][i*4 + 2] = colortable[n][intensity][2];
		    texvoldata[k][j][i*4 + 3] = colortable[n][intensity][3];

		}
	    }

       }

}

void cave_volrend::recalc(int n)
{
    int tID = CAVEUniqueIndex();

    glDeleteTextures(1 , &tex[tID][n]);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glBindTexture(GL_TEXTURE_3D, tex[tID][n]);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage3D(GL_TEXTURE_3D, 0, GL_RGBA, IV1, IV2, IV3,
		     0, GL_RGBA, GL_UNSIGNED_BYTE,
		     &texvoldata[0][0][0]);

  recalc_flag = 0;
}

//
//////////////////////////////////////
void cave_volrend::init_vertx()
{
/*

                      2_________ 3
                     /|        /| 
                    / |       / |    
                  0/_________/1 |     
                   |  |_____ |__|     
                   |  /6     |  /7
                   | /       | /              z  x
                   |/________|/               | /
                    4         5          y ___|/                   

  Origin is the center of data;

*/


    dvertx[0][0] = -datasize[0] / 2.0;
    dvertx[0][1] = datasize[1] / 2.0;
    dvertx[0][2] = datasize[2] / 2.0;

    dvertx[1][0] = -datasize[0] / 2.0;
    dvertx[1][1] = -datasize[1] / 2.0;
    dvertx[1][2] = datasize[2] / 2.0;

    dvertx[2][0] = datasize[0] / 2.0;
    dvertx[2][1] = datasize[1] / 2.0;
    dvertx[2][2] = datasize[2] / 2.0;

    dvertx[3][0] = datasize[0] / 2.0;
    dvertx[3][1] = -datasize[1] / 2.0;
    dvertx[3][2] = datasize[2] / 2.0;

    dvertx[4][0] = -datasize[0] / 2.0;
    dvertx[4][1] = datasize[1] / 2.0;
    dvertx[4][2] = -datasize[2] / 2.0;

    dvertx[5][0] = -datasize[0] / 2.0;
    dvertx[5][1] = -datasize[1] / 2.0;
    dvertx[5][2] = -datasize[2] / 2.0;

    dvertx[6][0] = datasize[0] / 2.0;
    dvertx[6][1] = datasize[1] / 2.0;
    dvertx[6][2] = -datasize[2] / 2.0;

    dvertx[7][0] = datasize[0] / 2.0;
    dvertx[7][1] = -datasize[1] / 2.0;
    dvertx[7][2] = -datasize[2] / 2.0;
}

////////////////////////////////////////
// This member function 
// is implemented by rendering processes
////////////////////////////////////////
void cave_volrend::draw()
{
    float scale;

    if (!on)
	return;

    switch (dispmode) {

    case 0:
	////////////////////
	// Volume Rendering
	////////////////////
	glDepthMask(GL_FALSE);
	draw_volumedata();
	glDepthMask(GL_TRUE);
	break;

    case 1:
	//////////////////
	// Color Editor
	//////////////////
	glPopMatrix();
	draw_coloreditor();
	glPushMatrix();
	CAVENavTransform();	// CAVE navigation

	scale = scalefh->current_scale_is();
	glScalef(scale, scale, scale);
	break;

      defalut:
	break;
    }
}


//////////////////////////////////////
//Resetting color table
//////////////////////////////////////
void cave_volrend::setcolortable()
{
    int i;

    float normed_v;

    float hue;
    float sat = 1.0;
    float val = 1.0;

    float r, g, b;


    for (i = 0; i < SCALAR_MAX; i++) {

	// set color table
	///////////////////
	normed_v = (255.0 - i) / 255.0;
	hue = 0.7 * normed_v;

	hsv2rgb(hue, sat, val, &r, &g, &b);

	for(int j=0;j<data->nscal;j++){
	colortable[j][i][0] = (unsigned char) (255.0 * r);
	colortable[j][i][1] = (unsigned char) (255.0 * g);
	colortable[j][i][2] = (unsigned char) (255.0 * b);
	// set alpha values
	////////////////////

	colortable[j][i][3] = 7;
	}

    }


}



//////////////////////////////////////
// Calculation of vertex of slices
//////////////////////////////////////
void cave_volrend::plane_calc()
{
    int i, j, slice_num = 0;
    int num_dvertx_above_slice, num_intersection, dvertx_above[8];
    float x0[3];
    float slc_plane, tmp_plane;
    float n[3], a;
    float head_pos[3], pos[3], head_direc[3];
    float temp, delta;
    float slice_vertx[6][7];	// 0,1,2(coordinate) 3,4,5(relative coordinate) 6(angle)
    float temp_slice_vertx[7];

    float local_origin[3];
    float vec0[3], magvec0, inprod;
    float scale;

    scale = scaleintp();

    if(dispmode != 0) return;

//    static int first =1;

    CAVEGetPosition(CAVE_HEAD_NAV, pos);


    head_pos[0] = pos[0] / scale - data_origin[0];
    head_pos[1] = pos[1] / scale - data_origin[1];
    head_pos[2] = pos[2] / scale - data_origin[2];


    /////////////////////////////////////////////////
    // Check if the viewer is inside the data or not.
    /////////////////////////////////////////////////

/*

                      2_________ 3
                     /|        /| 
                    / |       / |    
                  0/_________/1 |     
                   |  |_____ |__|     
                   |  /6     |  /7
                   | /       | /              z  x
                   |/________|/               | /
                    4         5          y ___|/                   

  Origin is the center of data;

*/

    if ((head_pos[0] < dvertx[2][0] && head_pos[0] > dvertx[0][0])
	&& (head_pos[1] < dvertx[0][1] && head_pos[1] > dvertx[1][1])
	&& (head_pos[2] < dvertx[0][2] && head_pos[2] > dvertx[4][2])) {
	/////////////////////////////////
	// The viewer is inside the data
	/////////////////////////////////

	CAVEGetVector(CAVE_HEAD_FRONT_NAV, head_direc);

//	bcopy(head_direc, n, sizeof(float) * 3);
	memcpy(n, head_direc, sizeof(float) * 3);

    } else {
	/////////////////////////////////////////
	// The viewer is outside the data.
	// The viewing direction is 
	// from the eye to the center of the data
	/////////////////////////////////////////
	n[0] = -head_pos[0];
	n[1] = -head_pos[1];
	n[2] = -head_pos[2];
    }


    temp = sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);

    n[0] /= temp;
    n[1] /= temp;
    n[2] /= temp;

    /////////////////////////////////////////////////////
    // Search the farthest plane which includes dvertx[i]
    // n[0]*x + n[1]*y+n[2]*z = XXX_plane
    /////////////////////////////////////////////////////
    int farthest_plane_num;
    float farthest_plane;
    float init_pos[3];

    farthest_plane_num = 0;
    farthest_plane =
	dvertx[0][0] * n[0] + dvertx[0][1] * n[1] + dvertx[0][2] * n[2];

    for (i = 1; i < 8; i++) {

	tmp_plane =
	    dvertx[i][0] * n[0] + dvertx[i][1] * n[1] +
	    dvertx[i][2] * n[2];

	if (tmp_plane > farthest_plane) {
	    farthest_plane = tmp_plane;
	    farthest_plane_num = i;
	}

    }

    init_pos[0] = dvertx[farthest_plane_num][0];
    init_pos[1] = dvertx[farthest_plane_num][1];
    init_pos[2] = dvertx[farthest_plane_num][2];

/*
	if(first){

	printf("init_pos:%f,%f,%f\n",init_pos[0],init_pos[1],init_pos[2]);	
	printf("region_max:%f,%f,%f\n",
	 data->region_xmax,data->region_ymax,data->region_zmax);	
	printf("region_min:%f,%f,%f\n",
	 data->region_xmin,data->region_ymin,data->region_zmin);	
	printf("data_origin:%f,%f,%f\n",
	 data_origin[0],data_origin[1],data_origin[2]);	


	first = 0;

	}
*/
    delta = slicedata.delta;
    a = delta;


    float dvtx_plane;

    while (1) {

	num_dvertx_above_slice = 0;
	num_intersection = 0;

	memset(dvertx_above, 0, sizeof(int) * 8);

	x0[0] = init_pos[0] - a * n[0];
	x0[1] = init_pos[1] - a * n[1];
	x0[2] = init_pos[2] - a * n[2];


	slc_plane = x0[0] * n[0] + x0[1] * n[1] + x0[2] * n[2];


	for (i = 0; i < 8; i++) {
	    dvtx_plane = n[0] * dvertx[i][0] + n[1] * dvertx[i][1] +
		n[2] * dvertx[i][2];

	    if (dvtx_plane > slc_plane) {

		dvertx_above[i] = 1;
		num_dvertx_above_slice++;

	    }

	}

	if (num_dvertx_above_slice == 8) {
	    // Exceeding nearest dvertex
	    break;
	}

	if (num_dvertx_above_slice == 0) {
	    a += delta;
	    continue;
	}


/*

                      2_________ 3
                     /|        /| 
                    / |       / |    
                  0/_________/1 |     
                   |  |_____ |__|     
                   |  /6     |  /7
                   | /       | /              z  x
                   |/________|/               | /
                    4         5          y ___|/                   

  Origin is the center of data;

*/

	if (dvertx_above[0] + dvertx_above[2] == 1) {
	    //////////////////////////////////////////////////////////
	    // Found an intersection point 
	    // between current slice and 0-2 line segment
	    // x = (plane - n[1]*y - n[2]*z)/n[0]
	    // if n[0] == 0 then there can't be no intersectional point
	    ///////////////////////////////////////////////////////////

	    slice_vertx[num_intersection][0] =
		(slc_plane - n[1] * dvertx[0][1] -
		 n[2] * dvertx[0][2]) / n[0];

	    slice_vertx[num_intersection][1] = dvertx[0][1];

	    slice_vertx[num_intersection][2] = dvertx[0][2];

	    num_intersection++;

	}


	if (dvertx_above[1] + dvertx_above[3] == 1) {

	    slice_vertx[num_intersection][0] =
		(slc_plane - n[1] * dvertx[1][1] -
		 n[2] * dvertx[1][2]) / n[0];

	    slice_vertx[num_intersection][1] = dvertx[1][1];

	    slice_vertx[num_intersection][2] = dvertx[1][2];

	    num_intersection++;

	}

	if (dvertx_above[4] + dvertx_above[6] == 1) {

	    slice_vertx[num_intersection][0] =
		(slc_plane - n[1] * dvertx[4][1] -
		 n[2] * dvertx[4][2]) / n[0];

	    slice_vertx[num_intersection][1] = dvertx[4][1];

	    slice_vertx[num_intersection][2] = dvertx[4][2];

	    num_intersection++;

	}

	if (dvertx_above[5] + dvertx_above[7] == 1) {

	    slice_vertx[num_intersection][0] =
		(slc_plane - n[1] * dvertx[5][1] -
		 n[2] * dvertx[5][2]) / n[0];

	    slice_vertx[num_intersection][1] = dvertx[5][1];

	    slice_vertx[num_intersection][2] = dvertx[5][2];

	    num_intersection++;

	}


	if (dvertx_above[0] + dvertx_above[1] == 1) {

	    slice_vertx[num_intersection][0] = dvertx[0][0];

	    slice_vertx[num_intersection][1] =
		(slc_plane - n[0] * dvertx[0][0] -
		 n[2] * dvertx[0][2]) / n[1];

	    slice_vertx[num_intersection][2] = dvertx[0][2];

	    num_intersection++;

	}

	if (dvertx_above[2] + dvertx_above[3] == 1) {

	    slice_vertx[num_intersection][0] = dvertx[2][0];

	    slice_vertx[num_intersection][1] =
		(slc_plane - n[0] * dvertx[2][0] -
		 n[2] * dvertx[2][2]) / n[1];

	    slice_vertx[num_intersection][2] = dvertx[2][2];

	    num_intersection++;

	}

	if (dvertx_above[4] + dvertx_above[5] == 1) {

	    slice_vertx[num_intersection][0] = dvertx[4][0];

	    slice_vertx[num_intersection][1] =
		(slc_plane - n[0] * dvertx[4][0] -
		 n[2] * dvertx[4][2]) / n[1];

	    slice_vertx[num_intersection][2] = dvertx[4][2];

	    num_intersection++;

	}


	if (dvertx_above[6] + dvertx_above[7] == 1) {

	    slice_vertx[num_intersection][0] = dvertx[6][0];

	    slice_vertx[num_intersection][1] =
		(slc_plane - n[0] * dvertx[6][0] -
		 n[2] * dvertx[6][2]) / n[1];

	    slice_vertx[num_intersection][2] = dvertx[6][2];

	    num_intersection++;

	}


	if (dvertx_above[0] + dvertx_above[4] == 1) {

	    slice_vertx[num_intersection][0] = dvertx[0][0];

	    slice_vertx[num_intersection][1] = dvertx[0][1];

	    slice_vertx[num_intersection][2] =
		(slc_plane - n[0] * dvertx[0][0] -
		 n[1] * dvertx[0][1]) / n[2];

	    num_intersection++;

	}

	if (dvertx_above[1] + dvertx_above[5] == 1) {

	    slice_vertx[num_intersection][0] = dvertx[1][0];

	    slice_vertx[num_intersection][1] = dvertx[1][1];

	    slice_vertx[num_intersection][2] =
		(slc_plane - n[0] * dvertx[1][0] -
		 n[1] * dvertx[1][1]) / n[2];

	    num_intersection++;

	}

	if (dvertx_above[3] + dvertx_above[7] == 1) {

	    slice_vertx[num_intersection][0] = dvertx[3][0];

	    slice_vertx[num_intersection][1] = dvertx[3][1];

	    slice_vertx[num_intersection][2] =
		(slc_plane - n[0] * dvertx[3][0] -
		 n[1] * dvertx[3][1]) / n[2];

	    num_intersection++;

	}

	if (dvertx_above[2] + dvertx_above[6] == 1) {

	    slice_vertx[num_intersection][0] = dvertx[2][0];

	    slice_vertx[num_intersection][1] = dvertx[2][1];

	    slice_vertx[num_intersection][2] =
		(slc_plane - n[0] * dvertx[2][0] -
		 n[1] * dvertx[2][1]) / n[2];

	    num_intersection++;

	}


	if (num_intersection < 3) {
	    puts("gosa");
	    a += delta;
	    continue;
	}


	slicedata.angles[slice_num] = num_intersection;


	//////////////////////////////////////////
	// Rearrange the vertex to draw a polygon 
	//////////////////////////////////////////
	if (num_intersection != 3) {

	    memset(local_origin, 0, sizeof(float) * 3);


	    local_origin[0] =
		slice_vertx[0][0] + (slice_vertx[1][0] -
				     slice_vertx[0][0]) / 2.0;
	    local_origin[1] =
		slice_vertx[0][1] + (slice_vertx[1][1] -
				     slice_vertx[0][1]) / 2.0;
	    local_origin[2] =
		slice_vertx[0][2] + (slice_vertx[1][2] -
				     slice_vertx[0][2]) / 2.0;


	    for (i = 0; i < num_intersection; i++) {

		slice_vertx[i][3] = slice_vertx[i][0] - local_origin[0];
		slice_vertx[i][4] = slice_vertx[i][1] - local_origin[1];
		slice_vertx[i][5] = slice_vertx[i][2] - local_origin[2];

	    }

	    // angle

	    slice_vertx[0][6] = 0.0;

	    vec0[0] = slice_vertx[0][3];
	    vec0[1] = slice_vertx[0][4];
	    vec0[2] = slice_vertx[0][5];

	    magvec0 =
		sqrt(vec0[0] * vec0[0] + vec0[1] * vec0[1] +
		     vec0[2] * vec0[2]);

	    vec0[0] /= magvec0;
	    vec0[1] /= magvec0;
	    vec0[2] /= magvec0;


	    for (i = 1; i < num_intersection; i++) {

		inprod =
		    slice_vertx[i][3] * vec0[0] +
		    slice_vertx[i][4] * vec0[1] +
		    slice_vertx[i][5] * vec0[2];

		temp =
		    sqrt(slice_vertx[i][3] * slice_vertx[i][3] +
			 slice_vertx[i][4] * slice_vertx[i][4] +
			 slice_vertx[i][5] * slice_vertx[i][5]);

		temp = inprod / temp;

		if (temp <= -1.0)
		    slice_vertx[i][6] = PI;
		else if (temp >= 1.0)
		    slice_vertx[i][6] = 0.0;
		else
		    slice_vertx[i][6] = acos(temp);

		temp =
		    (vec0[1] * slice_vertx[i][5] -
		     vec0[2] * slice_vertx[i][4]) * n[0]
		    + (vec0[2] * slice_vertx[i][3] -
		       vec0[0] * slice_vertx[i][5]) * n[1]
		    + (vec0[0] * slice_vertx[i][4] -
		       vec0[1] * slice_vertx[i][3]) * n[2];

		if (temp > 0.0)
		    slice_vertx[i][6] = 2.0 * PI - slice_vertx[i][6];

	    }

	    for (i = 1; i < num_intersection - 1; i++) {
		for (j = i + 1; j < num_intersection; j++) {

		    if (slice_vertx[i][6] < slice_vertx[j][6]) {

			memcpy(temp_slice_vertx,slice_vertx[i], 
			      sizeof(float) * 7);
			memcpy(slice_vertx[i],slice_vertx[j], 
			      sizeof(float) * 7);
			memcpy(slice_vertx[j],temp_slice_vertx, 
			      sizeof(float) * 7);

		    }


		}
	    }


	}			// if angle


	for (i = 0; i < num_intersection; i++) {
	    for (j = 0; j < 3; j++) {
		slicedata.slicevertx[slice_num].vertx[i][j] =
		    slice_vertx[i][j];
	    }
	}

	a += delta;
	slice_num++;

	if (slice_num >= slicedata.slice_max_limit)
	    break;

    }

    slicedata.num_slice = slice_num;

}

///////////////////////////////////////////////
// Move the ball by wand (Changing the curve)
///////////////////////////////////////////////
void cave_volrend::update_held()
{
    int h, v; //be careful
    float x, y;
    float wandPos[3], wandFront[3];

    CAVEGetPosition(CAVE_WAND, wandPos);
    CAVEGetVector(CAVE_WAND_FRONT, wandFront);

    float tmp;
    tmp = (G_ZORIGIN - wandPos[2]) / wandFront[2];
    x = wandPos[0] + wandFront[0] * tmp;
    y = wandPos[1] + wandFront[1] * tmp;

	if (cedata.grab_f) {
	//---------------------
	// "F"reely moving ball
	///////////////////////
	h = (x - G_XORIGIN) / GRAPH_SCALE - cedata.prev_h;

	if(h > 3) h = 3;
	 else if (h < -3) h=-3;

	if (h != 0)
	    cedata.hsphere = cedata.prev_h + h;

	if (cedata.hsphere < 0)
	    cedata.hsphere = 0;
	else if (cedata.hsphere > SCALAR_MAX - 1)
	    cedata.hsphere = SCALAR_MAX - 1;

	v = (y - G_YORIGIN) / GRAPH_SCALE - cedata.prev_v;

	if(v > 3) v = 3;
	 else if (v < -3) v=-3;

	if (v != 0)
	    cedata.vsphere = cedata.prev_v + v;

	if (cedata.vsphere > SCALAR_MAX - 1)
	    cedata.vsphere = SCALAR_MAX - 1;
	else if (cedata.vsphere < 0)
	    cedata.vsphere = 0;

	// for h = 0
	colortable[sscal_][cedata.hsphere][cedata.
				   element] =
	    (unsigned char) cedata.vsphere;

	for(int i=1;i <= abs(h);i++){

	 if(h > 0){

	colortable[sscal_][cedata.prev_h + i][cedata.element] =
	    (unsigned char)(cedata.prev_v + i*(cedata.vsphere - cedata.prev_v)/abs(h));

	 } else if(h < 0) {

	colortable[sscal_][cedata.prev_h - i][cedata.element] =
	    (unsigned char)(cedata.prev_v + i*(cedata.vsphere - cedata.prev_v)/abs(h));

	 }


	}

	cedata.prev_v = cedata.vsphere;
	cedata.prev_h = cedata.hsphere;

    }

}


////////////////////////////////////
// Drawing Color Editor
////////////////////////////////////
void cave_volrend::draw_coloreditor()
{
    int i;
    float wandPos[3], wandFront[3];
    float x, y;
    GLfloat Gcolor[5][4] = {
	{1.0, 0.0, 0.0, 1.0},	// R
	{0.0, 1.0, 0.0, 1.0},	// G
	{0.0, 0.0, 1.0, 1.0},	// B
	{0.0, 1.0, 1.0, 1.0},	// A
	{1.0, 1.0, 1.0, 1.0}	// White
    };

    CAVEGetPosition(CAVE_WAND, wandPos);
    CAVEGetVector(CAVE_WAND_FRONT, wandFront);

    /////////////////////////////////  
    // Cross Point on the front wall
    /////////////////////////////////
    x = wandPos[0] + wandFront[0] * (G_ZORIGIN -
				     wandPos[2]) / wandFront[2];
    y = wandPos[1] + wandFront[1] * (G_ZORIGIN -
				     wandPos[2]) / wandFront[2];


    glPushAttrib(GL_LIGHTING_BIT);
    //////////////////////////////
    // Drawing balls on Graph   
    //////////////////////////////
    int tID = CAVEUniqueIndex();

    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHT1);
    glMaterialfv(GL_FRONT, GL_AMBIENT, Gcolor[1]);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, Gcolor[2]);
    glPushMatrix();
    glTranslatef(G_XORIGIN, G_YORIGIN, G_ZORIGIN);
    glTranslatef(cedata.hsphere * GRAPH_SCALE,
		 colortable[sscal_][(cedata.hsphere)][cedata.
					      element] * GRAPH_SCALE, 0.0);
    gluSphere(sphereObj[tID], BALL_RAD, 16, 16);
    glPopMatrix();


    ////////////////////////
    // Graph's frame
    ////////////////////////

    glDisable(GL_LIGHT0);
    glDisable(GL_LIGHT1);

    glColor3f(1.0, 1.0, 1.0);

    glPushMatrix();
    glTranslatef(G_XORIGIN, G_YORIGIN, G_ZORIGIN);
    glCallList(volren_graph[tID]);
    glPopMatrix();

    glLineWidth(1.0);


    //////////////////////////
    // Drawing Lines on Graph
    //////////////////////////
    glColor3f(1.0, 1.0, 1.0);

    glPushMatrix();
    glTranslatef(G_XORIGIN, G_YORIGIN, G_ZORIGIN);

    glBegin(GL_LINE_STRIP);
    glVertex3f(0.0,
	       (colortable[sscal_][cedata.hsphere][cedata.element]) * GRAPH_SCALE,
	       0.0);
    glVertex3f(cedata.hsphere * GRAPH_SCALE,
	       colortable[sscal_][cedata.hsphere][cedata.element] * GRAPH_SCALE,
	       0.0);
    glVertex3f(cedata.hsphere * GRAPH_SCALE, 0.0, 0.0);
    glEnd();
    glPopMatrix();


    //////////////////////////////////////////////
    // Drawing Buttons for switching R G B A Reset
    //////////////////////////////////////////////
    glPushMatrix();
    glTranslatef(G_XORIGIN, G_YORIGIN + SQ_YOFFSET, G_ZORIGIN);

    for (i = 0; i < 5; i++) {

	glColor3fv(Gcolor[i]);

	glBegin(GL_QUADS);
	glVertex3f(0.0, 0.0, 0.0);
	glVertex3f(SQ_LEN, 0.0, 0.0);
	glVertex3f(SQ_LEN, SQ_LEN, 0.0);
	glVertex3f(0.0, SQ_LEN, 0.0);
	glEnd();

	glTranslatef(SQ_LEN + SQ_SPACE, 0.0, 0.0);

    }
    glPopMatrix();

    //////////////////////////
    // Drawing Curve on Graph
    //////////////////////////
    glPushMatrix();

    glLineWidth(2.0);
    glTranslatef(G_XORIGIN, G_YORIGIN, G_ZORIGIN);

    glBegin(GL_LINE_STRIP);

    glColor3fv(Gcolor[cedata.element]);

    for (i = 0; i < SCALAR_MAX; i++)
	glVertex3f(i * GRAPH_SCALE,
		   colortable[sscal_][i][cedata.element] * GRAPH_SCALE, 0.0);

    glEnd();

    glPopMatrix();

///////////////////////////////////////////////////
// Drawing Ray from wand to the ball or front wall
///////////////////////////////////////////////////
    glLineWidth(2.0);

    if (cedata.grab_f) {

	//////////////////////////////////////////////////////////////
	// if grabbing the ball for changing the color or alpha curve
	//////////////////////////////////////////////////////////////

	glPushMatrix();

	glBegin(GL_LINES);
	glVertex3fv(wandPos);
	glVertex3f(cedata.hsphere * GRAPH_SCALE + G_XORIGIN,
		   cedata.vsphere * GRAPH_SCALE + G_YORIGIN, G_ZORIGIN);
	glEnd();

	glPopMatrix();

    } else if (cedata.grab_f == 0) {

	////////////////////
	// Grabbing no ball
	////////////////////
	if (wandFront[2] < 0) {

	    glColor3f(1.0, 1.0, 1.0);

	    glPushMatrix();

	    glBegin(GL_LINES);
	    glVertex3fv(wandPos);
	    glVertex3f(x, y, G_ZORIGIN);
	    glEnd();

	    glPopMatrix();
	}

    }

    glLineWidth(1.0);

    /////////////////////////////////////
    // Drawing Color Bar below the Graph
    /////////////////////////////////////
    glColor3f(1.0, 1.0, 1.0);

    glPushMatrix();

    glTranslatef(G_XORIGIN, G_YORIGIN - COLOR_BAR_HEIGHT / 2.0, G_ZORIGIN);

    glBegin(GL_QUAD_STRIP);

    for (i = 0; i < SCALAR_MAX; i++) {
	glColor3ubv(colortable[sscal_][i]);

	glVertex3f(i * GRAPH_SCALE, -COLOR_BAR_HEIGHT / 2.0, 0.0);
	glVertex3f(i * GRAPH_SCALE, COLOR_BAR_HEIGHT / 2.0, 0.0);

    }

    glEnd();

    glPopMatrix();

    glPopAttrib();

}


///////////////////////////////////////////////
// Check if the ray penetrates the ball or not 
///////////////////////////////////////////////
void cave_volrend::check_grab()
{
    int i;
    float x, y;
    float wandPos[3], wandFront[3];

    if (!cedata.ray_exist)
	return;

    if (cedata.grab_f == 0) {

	CAVEGetPosition(CAVE_WAND, wandPos);
	CAVEGetVector(CAVE_WAND_FRONT, wandFront);

	/////////////////////////////////  
	// Cross Point on the front wall
	/////////////////////////////////
	x = wandPos[0] + wandFront[0] * (G_ZORIGIN -
					 wandPos[2]) / wandFront[2];
	y = wandPos[1] + wandFront[1] * (G_ZORIGIN -
					 wandPos[2]) / wandFront[2];


	if (fabs(x - G_XORIGIN - cedata.hsphere * GRAPH_SCALE)
		 < BALL_RAD
		 && fabs(y - G_YORIGIN -
			 colortable[sscal_][cedata.hsphere][cedata.element]
			 * GRAPH_SCALE) < BALL_RAD) {
	    //////////////////////////////////////
	    // the ball for changing the color map
	    //////////////////////////////////////
	    cedata.grab_f = 1;
	}
	//////////////////////////////////////////
	// chack if the ray penetrates the button
	//////////////////////////////////////////
	for (i = 0; i < 5; i++) {

	    if (x >= G_XORIGIN + (SQ_LEN + SQ_SPACE) * i &&
		x <= SQ_LEN * (i + 1) + SQ_SPACE * i + G_XORIGIN &&
		y >= SQ_YOFFSET + G_YORIGIN &&
		y <= SQ_YOFFSET + SQ_LEN + G_YORIGIN) {

		if (i < 4)
		    cedata.element = i;
		else {
		      setcolortable();
		      reset = 1;
		     }
		break;
	    }

	}


	cedata.vsphere = colortable[sscal_][cedata.hsphere][cedata.element];

    }				// Grab != 0

}

////////////////////////////////////////////
// Doing Volume Rendering using Texture-Map
////////////////////////////////////////////
void cave_volrend::draw_volumedata()
{
    int i, j;
    float x_offset, y_offset, z_offset;

    x_offset = datasize[0] / 2.0;
    y_offset = datasize[1] / 2.0;
    z_offset = datasize[2] / 2.0;


    glColor3f(1.0, 1.0, 1.0);

    glPushMatrix();
    glPushAttrib(GL_LIGHTING_BIT);
    glDisable(GL_LIGHT0);
    glDisable(GL_LIGHT1);

//    glColorTable(GL_TEXTURE_COLOR_TABLE_SGI, GL_RGBA, SCALAR_MAX, GL_RGBA,
//		 GL_UNSIGNED_BYTE, colortable);

 //   glEnable(GL_TEXTURE_COLOR_TABLE_SGI);


    glEnable(GL_TEXTURE_3D);

    ////////////////////////////////////////////////
    // The origin of the coordinates in slicevertex 
    // is the data_origin[3].
    // It is why this translatef is needed.
    //////////////////////////////////////////////// 
    glTranslatef(data_origin[0], data_origin[1], data_origin[2]);

    int tID = CAVEUniqueIndex();
    glBindTexture(GL_TEXTURE_3D, tex[tID][sscal_]);

    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);


    for (i = 0; i < slicedata.num_slice; i++) {

/*
	///////////////////////////
	// Drawing Frames of Slices
	// No need to do this
	///////////////////////////

	if(i%5 == 0) continue;

	glBegin(GL_LINE_LOOP);
	glColor3f(1.0, 0.0, 0.0);
	glLineWidth(2.0);
	for (j = 0; j < slicedata.angles[i]; j++) {
	    glVertex3fv(slicedata.slicevertx[i].vertx[j]);
	}
	glEnd();

	glBegin(GL_POLYGON);
	glColor3f(1.0, 1.0, 1.0);
	for (j = 0; j < slicedata.angles[i]; j++) {
	    glVertex3fv(slicedata.slicevertx[i].vertx[j]);
	}
	glEnd();

	///////////////////////////
	///////////////////////////
	continue;
	///////////////////////////
*/
	switch (slicedata.angles[i]) {

	case 3:
	    glBegin(GL_TRIANGLES);
	    for (j = 0; j < 3; j++) {

		glTexCoord3f((slicedata.slicevertx[i].vertx[j][0] +
			      x_offset) / datasize[0],
			     (slicedata.slicevertx[i].vertx[j][1] +
			      y_offset) / datasize[1],
			     (slicedata.slicevertx[i].vertx[j][2] +
			      z_offset) / datasize[2]);

		///////////////////////////////////////////
		// Texture Coordinates are all (0.0 - 1.0)
		// It is why ....+ X_offset)/datasize[X]
		///////////////////////////////////////////


		glVertex3fv(slicedata.slicevertx[i].vertx[j]);

	    }
	    glEnd();
	    break;

	case 4:
	    glBegin(GL_TRIANGLES);
	    // Triangle No.1
	    for (j = 0; j < 3; j++) {
		glTexCoord3f((slicedata.slicevertx[i].vertx[j][0] +
			      x_offset) / datasize[0],
			     (slicedata.slicevertx[i].vertx[j][1] +
			      y_offset) / datasize[1],
			     (slicedata.slicevertx[i].vertx[j][2] +
			      z_offset) / datasize[2]);

		glVertex3fv(slicedata.slicevertx[i].vertx[j]);
	    }

	    // Triangle No.2
	    glTexCoord3f((slicedata.slicevertx[i].vertx[0][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[0][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[0][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[0]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[2][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[2][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[2][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[2]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[3][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[3][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[3][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[3]);

	    glEnd();
	    break;

	case 5:
	    glBegin(GL_TRIANGLES);
	    // Triangle Mo.1
	    glTexCoord3f((slicedata.slicevertx[i].vertx[0][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[0][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[0][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[0]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[1][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[1][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[1][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[1]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[3][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[3][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[3][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[3]);

	    // Triangle No.2
	    glTexCoord3f((slicedata.slicevertx[i].vertx[0][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[0][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[0][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[0]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[3][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[3][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[3][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[3]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[4][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[4][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[4][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[4]);

	    // Triangle No.3
	    glTexCoord3f((slicedata.slicevertx[i].vertx[1][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[1][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[1][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[1]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[2][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[2][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[2][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[2]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[3][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[3][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[3][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[3]);
	    glEnd();
	    break;

	case 6:
	    glBegin(GL_TRIANGLES);
	    // Triangle No.1
	    glTexCoord3f((slicedata.slicevertx[i].vertx[0][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[0][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[0][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[0]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[1][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[1][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[1][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[1]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[3][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[3][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[3][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[3]);

	    // Triangle No.2
	    glTexCoord3f((slicedata.slicevertx[i].vertx[0][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[0][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[0][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[0]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[3][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[3][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[3][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[3]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[4][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[4][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[4][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[4]);

	    // Triangle No.3
	    glTexCoord3f((slicedata.slicevertx[i].vertx[1][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[1][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[1][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[1]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[2][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[2][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[2][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[2]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[3][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[3][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[3][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[3]);

	    // Triangle No.4
	    glTexCoord3f((slicedata.slicevertx[i].vertx[0][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[0][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[0][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[0]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[5][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[5][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[5][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[5]);


	    glTexCoord3f((slicedata.slicevertx[i].vertx[4][0] +
			  x_offset) / datasize[0],
			 (slicedata.slicevertx[i].vertx[4][1] +
			  y_offset) / datasize[1],
			 (slicedata.slicevertx[i].vertx[4][2] +
			  z_offset) / datasize[2]);

	    glVertex3fv(slicedata.slicevertx[i].vertx[4]);

	    glEnd();
	    break;

	default:
//        glBegin (GL_POLYGON);
//        for (j = 0; j < slicedata.angles[i]; j++)
//          {
//
//
//            glTexCoord3f ((slicedata.slicevertx[i].vertx[j][0] +
//                           x_offset) / datasize[0],
//                          (slicedata.slicevertx[i].vertx[j][1] +
//                           y_offset) / datasize[1],
//                          (slicedata.slicevertx[i].vertx[j][2] +
//                           z_offset) / datasize[2]);
//
//            glVertex3fv (slicedata.slicevertx[i].vertx[j]);
//          }
//        glEnd ();
	    break;
	}

    }


    glDisable(GL_TEXTURE_3D);
//    glDisable(GL_TEXTURE_COLOR_TABLE_SGI);

    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

    glDisable(GL_BLEND);
    glPopAttrib();
    glPopMatrix();

}

//
//////////////////////////////////////
void cave_volrend::button3_has_pressed()
{
    turn_off();
}


//
//////////////////////////////////////
void cave_volrend::button2_has_released()
{
    if(cedata.grab_f == 1 || reset == 1) recalc_flag = 1;
    dispmode = 0;
    cedata.ray_exist = 0;
    cedata.grab_f = 0;
    reset = 0;
}


//
//////////////////////////////////////
int cave_volrend::mode_out()
{
    return dispmode;
}


//
//////////////////////////////////////
int cave_volrend::ifon()
{
    return on;
}
